class ClientMessageType:
    EVENT = "EVENT"
    REQUEST = "REQ"
    CLOSE = "CLOSE"

class RelayMessageType:
    EVENT = "EVENT"
    NOTICE = "NOTICE"
    END_OF_STORED_EVENTS = "EOSE"

    @staticmethod
    def is_valid(type: str) -> bool:
        if type == RelayMessageType.EVENT or type == RelayMessageType.NOTICE or type == RelayMessageType.END_OF_STORED_EVENTS:
            return True
        return False


import json
from dataclasses import dataclass

from .filter import Filters
from .message_type import ClientMessageType

@dataclass
class Request:
    subscription_id: str
    filters: Filters

    def to_message(self) -> str:
        message = [ClientMessageType.REQUEST, self.subscription_id]
        message.extend(self.filters.to_json_array())
        return json.dumps(message)



import secrets
import base64
import secp256k1
from cffi import FFI
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
from hashlib import sha256

from .delegation import Delegation
from .event import EncryptedDirectMessage, Event, EventKind
from . import bech32


class PublicKey:
    def __init__(self, raw_bytes: bytes=None) -> None:
        self.raw_bytes = raw_bytes

    def bech32(self) -> str:
        converted_bits = bech32.convertbits(self.raw_bytes, 8, 5)
        return bech32.bech32_encode("npub", converted_bits, bech32.Encoding.BECH32)

    def hex(self) -> str:
        return self.raw_bytes.hex()

    def verify_signed_message_hash(self, hash: str, sig: str) -> bool:
        pk = secp256k1.PublicKey(b"\x02" + self.raw_bytes, True)
        return pk.schnorr_verify(bytes.fromhex(hash), bytes.fromhex(sig), None, True)

    @classmethod
    def from_npub(cls, npub: str):
        """ Load a PublicKey from its bech32/npub form """
        hrp, data, spec = bech32.bech32_decode(npub)
        raw_public_key = bech32.convertbits(data, 5, 8)[:-1]
        return cls(bytes(raw_public_key))


class PrivateKey:
    def __init__(self, raw_secret: bytes=None) -> None:
        if not raw_secret is None:
            self.raw_secret = raw_secret
        else:
            self.raw_secret = secrets.token_bytes(32)

        sk = secp256k1.PrivateKey(self.raw_secret)
        self.public_key = PublicKey(sk.pubkey.serialize()[1:])

    @classmethod
    def from_nsec(cls, nsec: str):
        """ Load a PrivateKey from its bech32/nsec form """
        hrp, data, spec = bech32.bech32_decode(nsec)
        raw_secret = bech32.convertbits(data, 5, 8)[:-1]
        return cls(bytes(raw_secret))

    def bech32(self) -> str:
        converted_bits = bech32.convertbits(self.raw_secret, 8, 5)
        return bech32.bech32_encode("nsec", converted_bits, bech32.Encoding.BECH32)

    def hex(self) -> str:
        return self.raw_secret.hex()

    def tweak_add(self, scalar: bytes) -> bytes:
        sk = secp256k1.PrivateKey(self.raw_secret)
        return sk.tweak_add(scalar)

    def compute_shared_secret(self, public_key_hex: str) -> bytes:
        pk = secp256k1.PublicKey(bytes.fromhex("02" + public_key_hex), True)
        return pk.ecdh(self.raw_secret, hashfn=copy_x)

    def encrypt_message(self, message: str, public_key_hex: str) -> str:
        padder = padding.PKCS7(128).padder()
        padded_data = padder.update(message.encode()) + padder.finalize()

        iv = secrets.token_bytes(16)
        cipher = Cipher(algorithms.AES(self.compute_shared_secret(public_key_hex)), modes.CBC(iv))

        encryptor = cipher.encryptor()
        encrypted_message = encryptor.update(padded_data) + encryptor.finalize()

        return f"{base64.b64encode(encrypted_message).decode()}?iv={base64.b64encode(iv).decode()}"
    
    def encrypt_dm(self, dm: EncryptedDirectMessage) -> None:
        dm.content = self.encrypt_message(message=dm.cleartext_content, public_key_hex=dm.recipient_pubkey)

    def decrypt_message(self, encoded_message: str, public_key_hex: str) -> str:
        encoded_data = encoded_message.split('?iv=')
        encoded_content, encoded_iv = encoded_data[0], encoded_data[1]

        iv = base64.b64decode(encoded_iv)
        cipher = Cipher(algorithms.AES(self.compute_shared_secret(public_key_hex)), modes.CBC(iv))
        encrypted_content = base64.b64decode(encoded_content)

        decryptor = cipher.decryptor()
        decrypted_message = decryptor.update(encrypted_content) + decryptor.finalize()

        unpadder = padding.PKCS7(128).unpadder()
        unpadded_data = unpadder.update(decrypted_message) + unpadder.finalize()

        return unpadded_data.decode()

    def sign_message_hash(self, hash: bytes) -> str:
        sk = secp256k1.PrivateKey(self.raw_secret)
        sig = sk.schnorr_sign(hash, None, raw=True)
        return sig.hex()

    def sign_event(self, event: Event) -> None:
        if event.kind == EventKind.ENCRYPTED_DIRECT_MESSAGE and event.content is None:
            self.encrypt_dm(event)
        if event.public_key is None:
            event.public_key = self.public_key.hex()
        event.signature = self.sign_message_hash(bytes.fromhex(event.id))

    def sign_delegation(self, delegation: Delegation) -> None:
        delegation.signature = self.sign_message_hash(sha256(delegation.delegation_token.encode()).digest())

    def __eq__(self, other):
        return self.raw_secret == other.raw_secret


def mine_vanity_key(prefix: str = None, suffix: str = None) -> PrivateKey:
    if prefix is None and suffix is None:
        raise ValueError("Expected at least one of 'prefix' or 'suffix' arguments")

    while True:
        sk = PrivateKey()
        if prefix is not None and not sk.public_key.bech32()[5:5+len(prefix)] == prefix:
            continue
        if suffix is not None and not sk.public_key.bech32()[-len(suffix):] == suffix:
            continue
        break

    return sk


ffi = FFI()
@ffi.callback("int (unsigned char *, const unsigned char *, const unsigned char *, void *)")
def copy_x(output, x32, y32, data):
    ffi.memmove(output, x32, 32)
    return 1



# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.0.3.dev10+g37cb66b.d20230505'
__version_tuple__ = version_tuple = (0, 0, 3, 'dev10', 'g37cb66b.d20230505')



from .filter import Filters

class Subscription:
    def __init__(self, id: str, filters: Filters=None) -> None:
        self.id = id
        self.filters = filters

    def to_json_object(self):
        return { 
            "id": self.id, 
            "filters": self.filters.to_json_array() 
        }



from collections import UserList
from typing import List

from .event import Event, EventKind



class Filter:
    """
    NIP-01 filtering.

    Explicitly supports "#e" and "#p" tag filters via `event_refs` and `pubkey_refs`.

    Arbitrary NIP-12 single-letter tag filters are also supported via `add_arbitrary_tag`.
    If a particular single-letter tag gains prominence, explicit support should be
    added. For example:
        # arbitrary tag
        filter.add_arbitrary_tag('t', [hashtags])
    
        # promoted to explicit support
        Filter(hashtag_refs=[hashtags])
    """
    def __init__(
            self, 
            event_ids: List[str] = None, 
            kinds: List[EventKind] = None, 
            authors: List[str] = None, 
            since: int = None, 
            until: int = None, 
            event_refs: List[str] = None,       # the "#e" attr; list of event ids referenced in an "e" tag
            pubkey_refs: List[str] = None,      # The "#p" attr; list of pubkeys referenced in a "p" tag
            limit: int = None) -> None:
        self.event_ids = event_ids
        self.kinds = kinds
        self.authors = authors
        self.since = since
        self.until = until
        self.event_refs = event_refs
        self.pubkey_refs = pubkey_refs
        self.limit = limit

        self.tags = {}
        if self.event_refs:
            self.add_arbitrary_tag('e', self.event_refs)
        if self.pubkey_refs:
            self.add_arbitrary_tag('p', self.pubkey_refs)


    def add_arbitrary_tag(self, tag: str, values: list):
        """
            Filter on any arbitrary tag with explicit handling for NIP-01 and NIP-12
            single-letter tags.
        """
         # NIP-01 'e' and 'p' tags and any NIP-12 single-letter tags must be prefixed with "#"
        tag_key = tag if len(tag) > 1 else f"#{tag}"
        self.tags[tag_key] = values


    def matches(self, event: Event) -> bool:
        if self.event_ids is not None and event.id not in self.event_ids:
            return False
        if self.kinds is not None and event.kind not in self.kinds:
            return False
        if self.authors is not None and event.public_key not in self.authors:
            return False
        if self.since is not None and event.created_at < self.since:
            return False
        if self.until is not None and event.created_at > self.until:
            return False
        if (self.event_refs is not None or self.pubkey_refs is not None) and len(event.tags) == 0:
            return False

        if self.tags:
            e_tag_identifiers = set([e_tag[0] for e_tag in event.tags])
            for f_tag, f_tag_values in self.tags.items():
                # Omit any NIP-01 or NIP-12 "#" chars on single-letter tags
                f_tag = f_tag.replace("#", "")

                if f_tag not in e_tag_identifiers:
                    # Event is missing a tag type that we're looking for
                    return False
                
                # Multiple values within f_tag_values are treated as OR search; an Event
                # needs to match only one.
                # Note: an Event could have multiple entries of the same tag type
                # (e.g. a reply to multiple people) so we have to check all of them.
                match_found = False
                for e_tag in event.tags:
                    if e_tag[0] == f_tag and e_tag[1] in f_tag_values:
                        match_found = True
                        break
                if not match_found:
                    return False

        return True


    def to_json_object(self) -> dict:
        res = {}
        if self.event_ids is not None:
            res["ids"] = self.event_ids
        if self.kinds is not None:   
            res["kinds"] = self.kinds
        if self.authors is not None:
            res["authors"] = self.authors
        if self.since is not None:
            res["since"] = self.since
        if self.until is not None:
            res["until"] = self.until
        if self.limit is not None:
            res["limit"] = self.limit
        if self.tags:
            res.update(self.tags)

        return res



class Filters(UserList):
    def __init__(self, initlist: "list[Filter]"=[]) -> None:
        super().__init__(initlist)
        self.data: "list[Filter]"

    def match(self, event: Event):
        for filter in self.data:
            if filter.matches(event):
                return True
        return False

    def to_json_array(self) -> list:
        return [filter.to_json_object() for filter in self.data]


import json
from queue import Queue
from threading import Lock
from .message_type import RelayMessageType
from .event import Event

class EventMessage:
    def __init__(self, event: Event, subscription_id: str, url: str) -> None:
        self.event = event
        self.subscription_id = subscription_id
        self.url = url

class NoticeMessage:
    def __init__(self, content: str, url: str) -> None:
        self.content = content
        self.url = url

class EndOfStoredEventsMessage:
    def __init__(self, subscription_id: str, url: str) -> None:
        self.subscription_id = subscription_id
        self.url = url

class MessagePool:
    def __init__(self) -> None:
        self.events: Queue[EventMessage] = Queue()
        self.notices: Queue[NoticeMessage] = Queue()
        self.eose_notices: Queue[EndOfStoredEventsMessage] = Queue()
        self._unique_events: set = set()
        self.lock: Lock = Lock()
    
    def add_message(self, message: str, url: str):
        self._process_message(message, url)

    def get_event(self):
        return self.events.get()

    def get_notice(self):
        return self.notices.get()

    def get_eose_notice(self):
        return self.eose_notices.get()

    def has_events(self):
        return self.events.qsize() > 0

    def has_notices(self):
        return self.notices.qsize() > 0

    def has_eose_notices(self):
        return self.eose_notices.qsize() > 0

    def _process_message(self, message: str, url: str):
        message_json = json.loads(message)
        message_type = message_json[0]
        if message_type == RelayMessageType.EVENT:
            subscription_id = message_json[1]
            e = message_json[2]
            event = Event(
                e["content"],
                e["pubkey"],
                e["created_at"],
                e["kind"],
                e["tags"],
                e["sig"],
            )
            with self.lock:
                if not event.id in self._unique_events:
                    self.events.put(EventMessage(event, subscription_id, url))
                    self._unique_events.add(event.id)
        elif message_type == RelayMessageType.NOTICE:
            self.notices.put(NoticeMessage(message_json[1], url))
        elif message_type == RelayMessageType.END_OF_STORED_EVENTS:
            self.eose_notices.put(EndOfStoredEventsMessage(message_json[1], url))





import json
import time
import threading
from dataclasses import dataclass
from threading import Lock

from .event import Event
from .filter import Filters
from .message_pool import MessagePool
from .message_type import ClientMessageType
from .relay import Relay, RelayPolicy, RelayProxyConnectionConfig
from .request import Request



class RelayException(Exception):
    pass



@dataclass
class RelayManager:
    def __post_init__(self):
        self.relays: dict[str, Relay] = {}
        self.message_pool: MessagePool = MessagePool()
        self.lock: Lock = Lock()

    def add_relay(
            self, 
            url: str, 
            policy: RelayPolicy = RelayPolicy(),
            ssl_options: dict = None,
            proxy_config: RelayProxyConnectionConfig = None):

        relay = Relay(url, self.message_pool, policy, ssl_options, proxy_config)

        with self.lock:
            self.relays[url] = relay

        threading.Thread(
            target=relay.connect,
            name=f"{relay.url}-thread"
        ).start()
        
        threading.Thread(
            target=relay.queue_worker,
            name=f"{relay.url}-queue",
            daemon=True
        ).start()

        time.sleep(1)

    def remove_relay(self, url: str):
        with self.lock:
            if url in self.relays:
                relay = self.relays.pop(url)
                relay.close()

    def add_subscription_on_relay(self, url: str, id: str, filters: Filters):
        with self.lock:
            if url in self.relays:
                relay = self.relays[url]
                if not relay.policy.should_read:
                    raise RelayException(f"Could not send request: {url} is not configured to read from")
                relay.add_subscription(id, filters)
                request = Request(id, filters)
                relay.publish(request.to_message())
            else:
                raise RelayException(f"Invalid relay url: no connection to {url}") 

    def add_subscription_on_all_relays(self, id: str, filters: Filters):
        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_read:
                    relay.add_subscription(id, filters)
                    request = Request(id, filters)
                    relay.publish(request.to_message())

    def close_subscription_on_relay(self, url: str, id: str):
        with self.lock:
            if url in self.relays:
                relay = self.relays[url]
                relay.close_subscription(id)
                relay.publish(json.dumps(["CLOSE", id]))
            else:
                raise RelayException(f"Invalid relay url: no connection to {url}")

    def close_subscription_on_all_relays(self, id: str):
        with self.lock:
            for relay in self.relays.values():
                relay.close_subscription(id)
                relay.publish(json.dumps(["CLOSE", id]))

    def close_all_relay_connections(self):
        with self.lock:
            for url in self.relays:
                relay = self.relays[url]
                relay.close()

    def publish_event(self, event: Event):
        """ Verifies that the Event is publishable before submitting it to relays """
        if event.signature is None:
            raise RelayException(f"Could not publish {event.id}: must be signed")

        if not event.verify():
            raise RelayException(f"Could not publish {event.id}: failed to verify signature {event.signature}")

        with self.lock:
            for relay in self.relays.values():
                if relay.policy.should_write:
                    relay.publish(event.to_message())



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

with open("README.md","r",encoding="utf-8-sig") as f:
    readme = f.read()

requirements = [
    "cffi>=1.15.0",
    "cryptography>=37.0.4",
    "pycparser>=2.21",
    "python-socks>=2.1.1",
    "secp256k1>=0.14.0",
    "websocket-client>=1.3.3"
]

setup(
    name="nostr-chatbot",
    version="0.1.0dev2",
    description="Nostr chatbotlibrary",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Wannaphong",
    author_email="wannaphong@yahoo.com",
    url="https://github.com/wannaphong/nostr-chatbot",
    packages=find_packages(),
    test_suite="tests",
    python_requires=">=3.8",
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords=[
        "chatbot",
        "nostr",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
    ],
)



import time
import json
from dataclasses import dataclass, field
from enum import IntEnum
from typing import List
from secp256k1 import PublicKey
from hashlib import sha256

from . import bech32
from .message_type import ClientMessageType



class EventKind(IntEnum):
    SET_METADATA = 0
    TEXT_NOTE = 1
    RECOMMEND_RELAY = 2
    CONTACTS = 3
    ENCRYPTED_DIRECT_MESSAGE = 4
    DELETE = 5



@dataclass
class Event:
    content: str = None
    public_key: str = None
    created_at: int = None
    kind: int = EventKind.TEXT_NOTE
    tags: List[List[str]] = field(default_factory=list)  # Dataclasses require special handling when the default value is a mutable type
    signature: str = None


    def __post_init__(self):
        if self.content is not None and not isinstance(self.content, str):
            # DMs initialize content to None but all other kinds should pass in a str
            raise TypeError("Argument 'content' must be of type str")

        if self.created_at is None:
            self.created_at = int(time.time())


    @staticmethod
    def serialize(public_key: str, created_at: int, kind: int, tags: List[List[str]], content: str) -> bytes:
        data = [0, public_key, created_at, kind, tags, content]
        data_str = json.dumps(data, separators=(',', ':'), ensure_ascii=False)
        return data_str.encode()


    @staticmethod
    def compute_id(public_key: str, created_at: int, kind: int, tags: List[List[str]], content: str):
        return sha256(Event.serialize(public_key, created_at, kind, tags, content)).hexdigest()


    @property
    def id(self) -> str:
        # Always recompute the id to reflect the up-to-date state of the Event
        return Event.compute_id(self.public_key, self.created_at, self.kind, self.tags, self.content)


    @property
    def note_id(self) -> str:
        converted_bits = bech32.convertbits(bytes.fromhex(self.id), 8, 5)
        return bech32.bech32_encode("note", converted_bits, bech32.Encoding.BECH32)


    def add_pubkey_ref(self, pubkey:str):
        """ Adds a reference to a pubkey as a 'p' tag """
        self.tags.append(['p', pubkey])


    def add_event_ref(self, event_id:str):
        """ Adds a reference to an event_id as an 'e' tag """
        self.tags.append(['e', event_id])


    def verify(self) -> bool:
        pub_key = PublicKey(bytes.fromhex("02" + self.public_key), True)  # add 02 for schnorr (bip340)
        return pub_key.schnorr_verify(bytes.fromhex(self.id), bytes.fromhex(self.signature), None, raw=True)


    def to_message(self) -> str:
        return json.dumps(
            [
                ClientMessageType.EVENT,
                {
                    "id": self.id,
                    "pubkey": self.public_key,
                    "created_at": self.created_at,
                    "kind": self.kind,
                    "tags": self.tags,
                    "content": self.content,
                    "sig": self.signature
                }
            ]
        )



@dataclass
class EncryptedDirectMessage(Event):
    recipient_pubkey: str = None
    cleartext_content: str = None
    reference_event_id: str = None


    def __post_init__(self):
        if self.content is not None:
            self.cleartext_content = self.content
            self.content = None

        if self.recipient_pubkey is None:
            raise Exception("Must specify a recipient_pubkey.")

        self.kind = EventKind.ENCRYPTED_DIRECT_MESSAGE
        super().__post_init__()

        # Must specify the DM recipient's pubkey in a 'p' tag
        self.add_pubkey_ref(self.recipient_pubkey)

        # Optionally specify a reference event (DM) this is a reply to
        if self.reference_event_id is not None:
            self.add_event_ref(self.reference_event_id)


    @property
    def id(self) -> str:
        if self.content is None:
            raise Exception("EncryptedDirectMessage `id` is undefined until its message is encrypted and stored in the `content` field")
        return super().id



import json
import time
from dataclasses import dataclass
from queue import Queue
from threading import Lock
from typing import Optional
from websocket import WebSocketApp
from .event import Event
from .filter import Filters
from .message_pool import MessagePool
from .message_type import RelayMessageType
from .subscription import Subscription

@dataclass
class RelayPolicy:
    should_read: bool = True
    should_write: bool = True
    
    def to_json_object(self) -> dict[str, bool]:
        return {"read": self.should_read, "write": self.should_write}




@dataclass
class RelayProxyConnectionConfig:
    host: Optional[str] = None
    port: Optional[int] = None
    type: Optional[str] = None



@dataclass
class Relay:
    url: str
    message_pool: MessagePool
    policy: RelayPolicy = RelayPolicy()
    ssl_options: Optional[dict] = None
    proxy_config: RelayProxyConnectionConfig = None

    def __post_init__(self):
        self.queue = Queue()
        self.subscriptions: dict[str, Subscription] = {}
        self.num_sent_events: int = 0
        self.connected: bool = False
        self.reconnect: bool = True
        self.error_counter: int = 0
        self.error_threshold: int = 0
        self.lock: Lock = Lock()
        self.ws: WebSocketApp = WebSocketApp(
            self.url,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close
        )

    def connect(self):
        self.ws.run_forever(
            sslopt=self.ssl_options,
            http_proxy_host=self.proxy_config.host if self.proxy_config is not None else None, 
            http_proxy_port=self.proxy_config.port if self.proxy_config is not None else None,
            proxy_type=self.proxy_config.type if self.proxy_config is not None else None,
        )

    def close(self):
        self.ws.close()

    def check_reconnect(self):
        try:
            self.close()
        except:
            pass
        self.connected = False
        if self.reconnect:
            time.sleep(1)
            self.connect()

    def publish(self, message: str):
        self.queue.put(message)
        
    def queue_worker(self):
        while True:
            if self.connected:
                message = self.queue.get()
                try:
                    self.ws.send(message)
                    self.num_sent_events += 1
                except:
                    self.queue.put(message)
            else:
                time.sleep(0.1)

    def add_subscription(self, id, filters: Filters):
        with self.lock:
            self.subscriptions[id] = Subscription(id, filters)

    def close_subscription(self, id: str) -> None:
        with self.lock:
            self.subscriptions.pop(id, None)

    def update_subscription(self, id: str, filters: Filters) -> None:
        with self.lock:
            subscription = self.subscriptions[id]
            subscription.filters = filters

    def to_json_object(self) -> dict:
        return {
            "url": self.url,
            "policy": self.policy.to_json_object(),
            "subscriptions": [
                subscription.to_json_object()
                for subscription in self.subscriptions.values()
            ],
        }

    def _on_open(self, class_obj):
        self.connected = True

    def _on_close(self, class_obj, status_code, message):
        self.connected = False

    def _on_message(self, class_obj, message: str):
        self.message_pool.add_message(message, self.url)
    
    def _on_error(self, class_obj, error):
        self.connected = False
        self.error_counter += 1
        if self.error_threshold and self.error_counter > self.error_threshold:
            pass
        else:
            self.check_reconnect()

    def _is_valid_message(self, message: str) -> bool:
        message = message.strip("\n")
        if not message or message[0] != "[" or message[-1] != "]":
            return False

        message_json = json.loads(message)
        message_type = message_json[0]
        if not RelayMessageType.is_valid(message_type):
            return False
        if message_type == RelayMessageType.EVENT:
            if not len(message_json) == 3:
                return False

            subscription_id = message_json[1]
            with self.lock:
                if subscription_id not in self.subscriptions:
                    return False

            e = message_json[2]
            event = Event(
                e["content"],
                e["pubkey"],
                e["created_at"],
                e["kind"],
                e["tags"],
                e["sig"],
            )
            if not event.verify():
                return False

            with self.lock:
                subscription = self.subscriptions[subscription_id]

            if subscription.filters and not subscription.filters.match(event):
                return False

        return True



import time
import uuid
import logging
from collections import defaultdict
from .nostr.key import PrivateKey,PublicKey
from .nostr.filter import Filter, Filters
from .nostr.event import Event, EventKind
from .nostr.relay_manager import RelayManager
from .nostr.message_type import ClientMessageType
from .nostr.event import EncryptedDirectMessage
from .nostr.relay_manager import RelayManager


class Chatbot:
    def __init__(
            self,
            private_key:str,
            recipient_pubkey:str,
            setup:bool=False,
            relay:list=["wss://nostr-pub.wellorder.net","wss://relay.damus.io","wss://nos.lol"]
        ) -> None:
        """
        :param str private_key: private_key (nsec)
        :param str recipient_pubkey: recipient public key (npub)
        :param bool setup: load history chat
        :param list relay: List relay
        """
        self.relay = relay
        if recipient_pubkey==None:
            self.recipient_pubkey = recipient_pubkey
        else:
            self.recipient_pubkey = PublicKey.from_npub(recipient_pubkey)
        self.private_key = PrivateKey.from_nsec(private_key)
        self.relay_manager = RelayManager()
        for i in self.relay:
            self.relay_manager.add_relay(i)
        self.subscription_id = uuid.uuid1().hex
        self.history = defaultdict(lambda: defaultdict(list))
        #{"bot":[],"recipient":[]}
        self.setup_receive()

        if setup:
            logging.info("setuping...")
            self.setup()
            logging.info("done!")
    
    def setup_receive(self):
        self.filters = Filters([
            Filter(
            kinds=[EventKind.ENCRYPTED_DIRECT_MESSAGE],
            pubkey_refs=[self.private_key.public_key.hex()]
            )
        ])
        self.relay_manager2 = RelayManager()
        for i in self.relay:
            self.relay_manager2.add_relay(i)
        self.subscription_id2 = uuid.uuid1().hex
        self.relay_manager2.add_subscription_on_all_relays(self.subscription_id2, self.filters)

    def setup(self):
        """
        for get old chat (from recipient response olny)
        """
        filters2 = Filters([
            Filter(
            kinds=[EventKind.ENCRYPTED_DIRECT_MESSAGE],
            authors=[self.private_key.public_key.hex()]
            )
        ])
        relay_manager3 = RelayManager()
        for i in self.relay:
            relay_manager3.add_relay(i)
        subscription_id3 = uuid.uuid1().hex
        relay_manager3.add_subscription_on_all_relays(subscription_id3, filters2)
        while relay_manager3.message_pool.has_events():
            event_msg = relay_manager3.message_pool.get_event()
            self.history[self.recipient_pubkey.bech32()]["bot"].append(event_msg.event)
        relay_manager3.close_subscription_on_all_relays(id=subscription_id3)
        while self.relay_manager2.message_pool.has_events():
            event_msg = self.relay_manager2.message_pool.get_event()
            self.history[self.recipient_pubkey.bech32()]["recipient"].append(event_msg.event)
        self.history[self.recipient_pubkey.bech32()]["bot"]=sorted(self.history[self.recipient_pubkey.bech32()]["bot"], key=lambda d: d.created_at)
        self.history[self.recipient_pubkey.bech32()]["recipient"]=sorted(self.history[self.recipient_pubkey.bech32()]["recipient"], key=lambda d: d.created_at)

    def send(self, message:str):
        """
        :param str message: message
        """
        dm = EncryptedDirectMessage(
            recipient_pubkey=self.recipient_pubkey.hex(),
            cleartext_content=message
        )
        self.private_key.sign_event(dm)
        self.relay_manager.publish_event(dm)
        self.history[self.recipient_pubkey.bech32()]["bot"].append(dm)

    def close(self):
        """
        close all subscription on all relays
        """
        self.relay_manager.close_subscription_on_all_relays(id=self.subscription_id)
        self.relay_manager2.close_subscription_on_all_relays(id=self.subscription_id2)

    def receive(self)->str:
        """
        receive a message from recipient.
        """
        while self.relay_manager2.message_pool.has_events():
            event_msg = self.relay_manager2.message_pool.get_event()
            if self.recipient_pubkey.bech32() not in self.history.keys():
                pass
            elif event_msg.event.created_at<self.history[self.recipient_pubkey.bech32()]["bot"][-1].created_at:
                return None
            self.history[self.recipient_pubkey.bech32()]["recipient"].append(event_msg.event)
        if self.history[self.recipient_pubkey.bech32()]["recipient"]==[]:
            return None
        if self.history[self.recipient_pubkey.bech32()]["recipient"][-1].created_at<self.history[self.recipient_pubkey.bech32()]["bot"][-1].created_at:
            return None
        e= self.history[self.recipient_pubkey.bech32()]["recipient"][-1]
        return self.private_key.decrypt_message(e.content,e.public_key)
